import pytest
import os
import subprocess as sb
import sys
from pypka.tests.builder.check_diffs import compareFiles

sys.path.insert(1, '../')

ncpus = 4

# run coverage.sh to generate coverage

def runTest(path, ncpus, results):
    os.system("rm -f {0}/*out {0}/clean*pqr {0}/TMP.gro {0}/delphi_in_*pdb".format(path))
    results_lines = results.split('\n')[1:-1]
    #"python3 -m coverage erase; "
    #"python3 -m coverage run ../../../pypka.py parameters.dat;"
    sb.Popen("cd {0}; "
             "sed -i 's/ncpus .*/ncpus           = {1}/' parameters.dat; "
             "python3 ../../../pypka.py parameters.dat;"
             "rm -f *-*.pdb *.profl *.prm *.xvg *.frc *.crg cent "
             "contributions interactions.dat pkint tmp.sites".format(path, ncpus), shell=True).wait()
    checkOutput('{0}/pKas.out'.format(path), results_lines)

def checkOutput(filename, results_lines):
    with open(filename) as f:
        c = -1
        for line in f:
            c += 1
            line = line.strip()
            assert line == results_lines[c]
    assert c + 1 == len(results_lines)

def checkStructureOutput(filename):
    problems = compareFiles(f'builder/{filename}', filename)
    if problems:
        raise Exception(f'Problems found with {filename}')
    else:
        os.remove(filename)


def checkAPIResult(pKa, results):
    i = -1
    for site in pKa:
        result = '{0} {1} {2} {3}'.format(site.res_name,
                                          site.res_number,
                                          site.pK,
                                          site.getProtState(7))
        i += 1
        assert result == results[i]
    assert i + 1 == len(results)

class TestCLI(object):
    def test_cli_ktp_gro(self):
        path = "ktp/ktp_gro"
        results = """
1 NTR 7.9006117525417885
1 TYR 9.88518370784027
2 CTR 3.123787446504994
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_allsites(self):
        path = "ktp/ktp_pdb_allsites"
        results = """
1 NTR 7.903609060282417
1 TYR 9.783802064791741
2 CTR 3.3021160843607587
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_allsites_noclean(self):
        path = "ktp/ktp_pdb_allsites_noclean"
        results = """
1 NTR 7.9006117525417885
1 TYR 9.88518370784027
2 CTR 3.123787446504994
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_onlytermini(self):
        path = 'ktp/ktp_pdb_onlytermini'
        results = """
1 NTR 7.900161128244222
2 CTR 3.2974246565617125
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_onlytermini_noclean(self):
        path = 'ktp/ktp_pdb_onlytermini_noclean'
        results = """
1 NTR 8.167721738527177
2 CTR 3.390519247038917
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_gro(self):
        path = "lyso/lyso_gro"
        results = """
1 NTR 7.636678712935952
18 ASP 3.140817760292259
35 GLU 4.663003222341569
48 ASP 2.964461211978307
66 ASP 3.255524771073022
129 CTR 2.3530479084733646
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_sites(self):
        path = "lyso/lyso_pdb_sites"
        results = """
1 NTR 7.506180977656024
18 ASP 3.1878485151212437
35 GLU 4.608229404531882
48 ASP 2.2579192691339616
66 ASP 1.820122723288205
129 CTR 2.3052342318764727
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_sites_noclean(self):
        path = "lyso/lyso_pdb_sites_noclean"
        results = """
1 NTR 7.393383742911153
18 ASP 2.7812887935898885
35 GLU 4.271359971006614
48 ASP 2.6547679911654636
66 ASP 3.010625039606261
129 CTR 1.9407914909927175
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_all(self):
        path = "lyso/lyso_pdb_all"
        results = """
1 NTR 7.385486289829849
1 LYS 10.385584365325077
7 GLU 2.8826946555866253
13 LYS None
15 HIS 5.867371279679988
18 ASP 2.699499030381383
20 TYR 9.911190805007351
23 TYR 9.392016503352243
24 SER None
33 LYS 10.512639059944679
35 GLU 4.175352330803779
36 SER None
40 THR None
43 THR None
47 THR None
48 ASP 1.859099446300949
50 SER None
51 THR None
52 ASP 2.1516355521689436
53 TYR 10.738445987683207
60 SER None
66 ASP 2.337319641232289
69 THR None
72 SER None
81 SER None
85 SER None
86 SER None
87 ASP 2.0567544019613804
89 THR None
91 SER None
96 LYS None
97 LYS None
100 SER None
101 ASP 3.31453580803786
116 LYS 10.253852282222178
118 THR None
119 ASP 2.3967126151939904
129 CTR 1.7030741123202797
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_all_noclean(self):
        path = "lyso/lyso_pdb_all_noclean"
        results = """
1 NTR 7.384356794782545
1 LYS 10.389604559472225
7 GLU 2.883256437556686
13 LYS None
15 HIS 5.857142857142857
18 ASP 2.704063319217584
20 TYR 9.912713556689594
23 TYR 9.394213400114952
24 SER None
33 LYS 10.516970339406788
35 GLU 4.177328004243067
36 SER None
40 THR None
43 THR None
47 THR None
48 ASP 1.8664671985815602
50 SER None
51 THR None
52 ASP 2.1504670878557155
53 TYR 10.748483872190425
60 SER None
66 ASP 2.3538356252272292
69 THR None
72 SER None
81 SER None
85 SER None
86 SER None
87 ASP 2.047742733457019
89 THR None
91 SER None
96 LYS None
97 LYS None
100 SER None
101 ASP 3.3119866025485094
116 LYS 10.254489298936823
118 THR None
119 ASP 2.397993604492454
129 CTR 1.702982198203926
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_multichain_all(self):
        path = "lyso/lyso_multichain_all"
        results = """
1 NTR 7.183462163467873
1 LYS 10.396268714909544
7 GLU 2.7881048585788903
13 NTR 2.4753604406285437
13 LYS None
15 HIS 5.766867861454174
18 ASP 2.409258450943305
20 CTR None
20 TYR 9.978007803338805
21 NTR None
23 TYR 8.51584948914053
24 SER None
33 LYS 10.516964051164953
35 GLU 4.053515554030321
36 SER None
40 THR None
41 NTR 6.30871506642378
43 THR None
47 THR None
48 ASP 1.8451897981859995
50 SER None
51 THR None
52 ASP 2.002075936318864
53 TYR 10.705104182281517
60 SER None
66 ASP 2.2240888633956937
69 THR None
72 SER None
81 SER None
85 SER None
86 SER None
87 ASP 1.891905748004584
89 THR None
91 SER None
96 LYS None
97 LYS None
100 SER None
101 ASP 3.338910637657156
116 LYS 10.240017754608644
118 THR None
119 ASP 2.355701271398538
129 CTR 1.8967363230805352
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_multichain_sites(self):
        path = "lyso/lyso_multichain_sites"
        results = """
1 NTR 7.2083941115933605
1 ASP 3.0432687937460243
35 GLU 4.612318525602017
35 ASP 2.2543511178409195
66 ASP 1.8262090921859637
1 CTR 2.702574220217964
        """
        runTest(path, ncpus, results)

    def test_cli_pHLIP_gro(self):
        path = "pHLIP/pHLIP_gro"
        results = """
769 NTR 10.7649336974175
770 CYS None
771 GLU 3.997796390277407
782 ASP None
793 ASP None
799 ASP 1.9435900082576383
801 ASP 0.9864795312376277
802 GLU 4.057797348110034
804 CTR 3.837918510075153
        """
        runTest(path, ncpus, results)

    def test_cli_pHLIP_pdb_all(self):
        path = "pHLIP/pHLIP_pdb_all"
        results = """
769 NTR 11.20572255781535
770 CYS 24.212240582368022
771 GLU 4.006244338659009
776 TYR 18.59753191489362
780 TYR 24.40723419370116
782 ASP None
786 THR None
787 THR None
793 ASP 22.721932396697188
799 ASP 0.4165757481296758
801 ASP 1.4362119419712347
802 GLU 3.945716473628793
804 CTR 4.773451582666378
804 THR None
        """
        runTest(path, ncpus, results)

    def test_cli_nucleosome_pdb_all(self):
        path = "nucleosome/nucleosome_pdb_all"
        results = """
1 NTR 7.882393397524072
3 THR None
4 LYS 10.290327706758951
6 THR None
9 LYS 8.516685074155886
10 SER None
11 THR None
14 LYS 10.06376595512564
18 LYS 7.63923256980438
22 THR None
23 LYS 9.672642410098653
27 LYS 10.839709485474273
28 SER None
32 THR None
36 LYS 10.414026770775237
37 LYS 11.254282087233317
39 HIS 8.760644085644085
41 TYR 13.72589107854297
45 THR None
50 GLU 3.0977474008471315
54 TYR 10.836061671427526
56 LYS 10.461906877554647
57 SER None
58 THR None
59 GLU 2.972853139247919
64 LYS 10.749247888651274
73 GLU 3.4225222172013856
77 ASP 4.501544770250534
79 LYS 10.883174614330393
80 THR None
81 ASP 2.58587883648028
86 SER None
87 SER None
94 GLU 2.336822582673205
96 SER None
97 GLU 0.2736797240504321
99 TYR 12.748583252553873
105 GLU 4.3537843607220434
106 ASP 2.758237453090849
107 THR None
110 CYS 13.30342405281947
113 HIS 3.462506496399139
115 LYS 10.421432084021216
118 THR None
122 LYS 8.99167708946439
123 ASP None
133 GLU 0.7110435350307528
135 CTR 3.5171617220137366
"""
        runTest(path, ncpus, results)


class TestAPI(object):
    def test_api_ktp_gro(self):
        from ..pypka import Titration
        os.system('rm -f LOG* *out *gro *pdb *pqr *crg *sites cent contributions interactions.dat pkint')
        results = """
NTR 2001 7.9006117525417885 ('undefined', 0.8883240472631488)
TYR 1 9.88518370784027 ('protonated', 0.99869907894847)
CTR 2002 3.123787446504994 ('deprotonated', 0.0001329626608755893)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_gro/ktp.gro',     # MANDATORY
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '-5,15',
            'pHstep'        : 0.2,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 500,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out',
            'clean_pdb'     : False
        }
        sites = {'A': ('1N', '1', '2C')}
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_allsites_clean(self):
        from ..pypka import Titration
        os.system('rm -f *out *gro *pdb *pqr *crg *sites cent contributions interactions.dat pkint')
        results = """
NTR 2001 7.903967864316001 ('undefined', 0.8890883751899552)
TYR 1 9.781315673289184 ('protonated', 0.9983481661893414)
CTR 2002 3.3021115745568306 ('deprotonated', 0.0002004585145523701)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_allsites/ktp.pdb',
            'clean_pdb'     : 'yes',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,
            'ncpus'         : ncpus,
            'pH'            : '0,15',
            'pHstep'        : 0.2,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'        : 81,
            'convergence'   : 0.01,
            'nlit'          : 300,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out',
            'clean_pdb'     : True
        }
        sites = 'all'
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_allsites_noclean(self):
        from ..pypka import Titration
        os.system('rm -rf LOG* *out *gro *pdb *pqr *crg *sites '
                  'cent contributions interactions.dat pkint LOG* __pycache__ *pyc')
        results = """
NTR 2001 7.9006117525417885 ('undefined', 0.8883240472631488)
TYR 1 9.88518370784027 ('protonated', 0.99869907894847)
CTR 2002 3.123787446504994 ('deprotonated', 0.0001329626608755893)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_allsites_noclean/ktp_noclean.pdb',     # MANDATORY
            'clean_pdb'     : 'no',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '-5,15',
            'pHstep'        : 0.2,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 500,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out',
            'clean_pdb'     : False
        }
        sites = 'all'
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_onlytermini(self):
        from ..pypka import Titration
        os.system('rm -rf LOG* *out *gro *pdb *pqr *crg *sites cent '
                  'contributions interactions.dat pkint LOG* __pycache__ *pyc')
        results = """
NTR 2001 7.900161128244222 ('undefined', 0.8882210711812845)
CTR 2002 3.2974246565617125 ('deprotonated', 0.000198307219057625)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_onlytermini/ktp.pdb',     # MANDATORY
            'clean_pdb'     : 'yes',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '0,15',
            'pHstep'        : 0.25,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 300,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out',
            'clean_pdb'     : True
        }
        sites = {' ': ('1N', '2C')}
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_onlytermini_noclean(self):
        from ..pypka import Titration
        os.system('rm -rf LOG* *out *gro *pdb *pqr *crg *sites cent '
                  'contributions interactions.dat pkint LOG* __pycache__ *pyc')
        results = """
NTR 2001 8.167721738527177 ('protonated', 0.9363612426447844)
CTR 2002 3.390519247038917 ('deprotonated', 0.0002457041692844058)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_onlytermini_noclean/ktp_noclean.pdb',     # MANDATORY
            'clean_pdb'     : 'no',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '0,15',
            'pHstep'        : 0.25,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 300,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out',
            'clean_pdb'     : False
        }
        sites = {' ': ('1N', '2C')}
        pKa = Titration(parameters, sites=sites)

        checkAPIResult(pKa, results)


class TestBuilder(object):
    def test_lyso_pH0(self):
        outfile = 'amber1.pdb'
        from ..pypka import Titration

        parameters = {'structure'     : 'builder/4lzt.pdb',
                      'epsin'         : 15,
                      'ionicstr'      : 0.1,
                      'pbc_dimensions': 0,
                      'ncpus'         : ncpus,
                      'output'        : 'pKas.out',
                      'titration_output': 'titration.out',
                      'structure_output': (outfile, 1, "amber")
                      }
        pKa = Titration(parameters)
        checkStructureOutput(outfile)

    def test_lyso_pH7(self):
        outfile = 'amber7.pdb'
        from ..pypka import Titration

        parameters = {'structure'     : 'builder/4lzt.pdb',
                      'epsin'         : 15,
                      'ionicstr'      : 0.1,
                      'pbc_dimensions': 0,
                      'ncpus'         : ncpus,
                      'output'        : 'pKas.out',
                      'titration_output': 'titration.out',
                      'structure_output': (outfile, 7, "amber")
                      }
        pKa = Titration(parameters)
        checkStructureOutput(outfile)

    def test_lyso_pH12(self):
        outfile = 'amber12.pdb'
        from ..pypka import Titration

        parameters = {'structure'     : 'builder/4lzt.pdb',
                      'epsin'         : 15,
                      'ionicstr'      : 0.1,
                      'pbc_dimensions': 0,
                      'ncpus'         : ncpus,
                      'output'        : 'pKas.out',
                      'titration_output': 'titration.out',
                      'structure_output': (outfile, 12, "amber")
                      }
        pKa = Titration(parameters)
        checkStructureOutput(outfile)

    def test_lyso_pH12_gromos(self):
        outfile = 'gromos12.pdb'
        from ..pypka import Titration

        parameters = {'structure'     : 'builder/4lzt.pdb',
                      'epsin'         : 15,
                      'ionicstr'      : 0.1,
                      'pbc_dimensions': 0,
                      'ncpus'         : ncpus,
                      'output'        : 'pKas.out',
                      'titration_output': 'titration.out',
                      'structure_output': (outfile, 12, "gromos_CpH")
                      }
        pKa = Titration(parameters)
        checkStructureOutput(outfile)



""""Constants file
"""

# Constants
KBOLTZ = 5.98435e-6  # e^2/(Angstrom*K)
LOG10 = 2.302585092994046

# Lipids
LIPIDS = {'cholesterol': 'CHO',  # to edit
          'POPC': 'POP'}
LIPID_RESIDUES = ['DMX', 'POX', 'PJ2', 'PJ1', 'PJ0', 'CHL']  # allowed residue names

# Ions
IONS = ['CL-', 'NA+']

# Tautomers and residues
TERMINAL_OFFSET = 2000

TITRABLETAUTOMERS = {'LYS': 3,
                     'HIS': 2,
                     'ASP': 4,
                     'GLU': 4,
                     'SER': 3,
                     'THR': 3,
                     'CYS': 3,
                     'CTR': 4,
                     'NTR': 3,
                     'TYR': 2}

PROTEIN_RESIDUES = ('ALA', 'VAL', 'ILE', 'LEU', 'MET', 'PHE', 'TYR', 'TRP',
                    'ASP', 'GLU', 'ARG', 'HIS', 'LYS', 'SER', 'THR', 'ASN',
                    'GLN', 'CYS', 'NTR', 'CTR')

TITRABLERESIDUES = list(TITRABLETAUTOMERS.keys())
REGULARTITRATINGRES = list(TITRABLETAUTOMERS.keys())

for res in REGULARTITRATINGRES:
    ntautomers = TITRABLETAUTOMERS[res]
    for i in range(ntautomers + 1):
        TITRABLERESIDUES.append(res[0:2] + str(i))

MANDATORY_PARAMS = [
    'structure', 'epsin', 'ionicstr',
    'pbc_dimensions', 'ncpus']

IGNORED_PARAMS = [
    'tmpsites', 'pid', 'file_dir', 'script_dir',
    'pdb2pqr', 'userff', 'usernames',
    'delphi_params', 'mc_params', 'pH', 'lipid_definition', 'f_dat']

MAXNPKHALFS = 3
PKAPLACEHOLDER = None




import os
import log
from formats import (read_pdb_line, read_pqr_line, read_gro_line,
                     correct_names, new_pqr_line, new_pdb_line, new_gro_line, pdb2gro)

from config import Config
from constants import *
from ffconverter import *
from copy import copy

def inputPDBCheck(filename, sites, clean_pdb):
    """
    Returns: chains_length, chains_res
    """
    if filename[-3:] in ('pdb', 'pqr'):
        filetype = 'pdb'
    elif filename[-3:] == 'gro':
        filetype = 'gro'
    else:
        raise Exception('Input file must be either a pdb or a gro.')

    chains_length = {}
    chains_res = {}

    for chain in sites.keys():
        chains_res[chain] = {}
        for site in sites[chain]:
            if site[-1] == 'C':
                resnumb = site[:-1]
                chains_res[chain][resnumb] = 'CTR'
            elif site[-1] == 'N':
                resnumb = site[:-1]
                chains_res[chain][resnumb] = 'NTR'

    if filetype == 'pdb' and not clean_pdb:
        new_gro_header = 'CREATED within PyPka\n'
        new_gro_body = ''
    with open(filename) as f:
        last_chain = ''
        chain_length = 0

        nline = 0
        maxnlines = 0
        atom_number = 0
        for line in f:
            nline += 1
            atom_line = False
            if filetype == 'pdb':
                if 'ATOM ' == line[0:5]:
                    atom_line = True
                    chain_length += 1
                    (aname, anumb, resname,
                     chain, resnumb, x, y, z) = read_pdb_line(line)
                    atom_number += 1
                    if not clean_pdb:
                        if len(aname) > 2 and \
                           aname[1] == 'H' and \
                           aname[0] in ('1', '2'):
                            aname = aname[1:] + aname[0]
                        new_gro_body += new_gro_line(anumb, aname,
                                                     resname, resnumb,
                                                     x / 10.0, y / 10, z / 10)
                elif 'CRYST1' in line:
                    tmp = line.split()[1:4]
                    box = (float(tmp[0]), float(tmp[1]), float(tmp[2]))
                    new_gro_footer = '{0:10.5f}{1:10.5f}{2:10.5f}\n'.format(box[0] / 10.0,
                                                                            box[1] / 10.0,
                                                                            box[2] / 10.0)

            elif filetype == 'gro':
                if nline > 2 and nline < maxnlines:
                    (aname, anumb, resname, resnumb, x, y, z) = read_gro_line(line)
                    chain = 'A'
                    atom_line = True
                elif nline == 2:
                    natoms = int(line.strip())
                    maxnlines = natoms + 3

            if atom_line:
                if chain_length == 1:
                    last_chain = chain

                if chain != last_chain and chain_length != 1:
                    chains_length[last_chain] = chain_length
                    #chains_res[chain] = done[chain]
                    chain_length = 0
                    last_chain = chain

                if chain in sites and \
                   resnumb not in chains_res[chain] and \
                   str(resnumb) in sites[chain]:
                    chains_res[chain][resnumb] = resname

    #if filetype == 'pdb' and not clean_pdb:
    #    new_gro_header += '{0}\n'.format(atom_number)
    #    with open('TMP.gro', 'w') as f:
    #        f.write(new_gro_header + new_gro_body + new_gro_footer)

    chains_length[last_chain] = chain_length
    #chains_res[chain] = done[chain]

    return chains_length, chains_res


def cleanPDB(molecules, chains_res, inputpqr, outputpqr):
    """
    """
    pdb_filename = Config.pypka_params['f_in']
    pdb2pqr_path = Config.pypka_params['pdb2pqr']
    inputpdbfile = removeMembrane(pdb_filename)

    logfile = 'LOG_pdb2pqr'
    #errfile = 'LOG_pdb2pqr_err'

    #log.redirectOutput("start", logfile)
    #log.redirectErr("start", errfile)

    # CTR O1/O2 will be deleted and a O/OXT will be added
    os.system('python2 {0} {1} {2} --ff {4} --ffout {4} '
              '--drop-water -v --chain > {5} 2>&1 '.format(pdb2pqr_path,
                                                           inputpdbfile, inputpqr,
                                                           Config.pypka_params['ffinput'],
                                                           Config.pypka_params['ff_family'],
                                                           logfile))
    if Config.pypka_params['f_structure_out']:
        ff_out = Config.pypka_params['ff_structure_out']
        if ff_out == 'gromos_cph':
            ff_out = 'gromos'
        os.system('python2 {0} {1} Hs.pqr --ff {3} --ffout {3} '
                  '--drop-water -v --chain >> {2} 2>&1 '.format(pdb2pqr_path,
                                                                inputpdbfile,
                                                                logfile,
                                                                ff_out))

    #log.redirectOutput("stop", logfile)
    #log.redirectErr("stop", errfile)

    if Config.pypka_params['f_structure_out']:
        with open('Hs.pqr') as f:
            for line in f:
                if line.startswith('ATOM '):
                    (aname, anumb, resname, chain, resnumb, x, y,
                     z, charge, radius) = read_pqr_line(line)

                    if resnumb in chains_res['A'] and \
                       resname in AMBER_Hs and \
                       aname in AMBER_Hs[resname] and \
                       aname not in AMBER_mainchain_Hs:
                        continue
                    elif aname[0] == 'H' and aname not in ('H1', 'H2', 'H3'):
                        if not resnumb in mainchain_Hs:
                            mainchain_Hs[resnumb] = []
                        mainchain_Hs[resnumb].append((aname, anumb, resname, chain,
                                                      x, y, z))

    CYS_bridges = {}
    with open('LOG_pdb2pqr') as f:
        for line in f:
            if 'patched with CYX' in line:
                parts = line.split('patched')[0].replace('PATCH INFO: ', '').split()
                resname, chain, resnumb = parts
                chain = chain.replace('_', ' ')
                if not chain in CYS_bridges:
                    CYS_bridges[chain] = []
                cys_res_numb = int(parts[-1])
                if cys_res_numb not in CYS_bridges[chain]:
                    CYS_bridges[chain].append(cys_res_numb)

    for chain, molecule in molecules.items():
        if chain in CYS_bridges:
            molecule.saveCYSBridges(CYS_bridges[chain])

    new_pdb_text = ''
    removed_pdb_text = ''
    removed_pdb_lines = []
    #ntr_trigger = True
    resnumb_max = 0
    chains = list(molecules.keys())

    with open(inputpqr) as f:
        for line in f:
            if "ATOM" in line[:4]:
                (aname, anumb, resname, original_chain, resnumb, x, y,
                 z, charge, radius) = read_pqr_line(line)

                if original_chain == '_':
                    chain = ' '
                else:
                    chain = original_chain

                termini_trigger = False
                if chain in chains:
                    molecule = molecules[chain]
                    NTR_numb = molecule.NTR
                    CTR_numb = molecule.CTR

                    sites_numbs = molecule.sites.keys()

                    aname, resname = correct_names(resnumb,
                                                   resname, aname,
                                                   sites_numbs, NTR_numb, CTR_numb)
                    resnumb_max = resnumb

                    if resnumb in (NTR_numb, CTR_numb):
                        termini_trigger = True

                if aname in ('O1', 'O2',
                             'OT1', 'OT2',
                             'H1', 'H2', 'H3') and \
                   not termini_trigger and resname in PROTEIN_RESIDUES:
                    if aname == 'O1':
                        aname = 'O'
                    elif aname == 'H1':
                        aname = 'H'
                    else:
                        continue

                if line[26] != ' ':
                    resnumb += TERMINAL_OFFSET

                new_line = new_pqr_line(anumb, aname, resname,
                                        resnumb, x, y, z, charge, radius, chain=original_chain)
                if chain in molecules:
                    new_pdb_text += new_line
                elif aname not in ('O1', 'O2',
                                   'OT1', 'OT2',
                                   'H1', 'H2', 'H3'):
                    if resname in ('SER', 'THR') and aname == 'HG1':
                        aname = 'HG'
                    removed_pdb_lines.append(new_line)

    resnumb_old = resnumb_max + 1
    for line in removed_pdb_lines:
        (aname, anumb_old, resname, chain, resnumb, x, y,
         z, charge, radius) = read_pqr_line(line)
        anumb += 1
        resnumb += resnumb_old
        while resnumb < resnumb_max:
            resnumb += resnumb_old
        removed_pdb_text += new_pqr_line(anumb, aname, resname,
                                         resnumb, x, y, z, charge, radius)
        resnumb_max = resnumb
    #            if ntr_trigger and :
    #                config.tit_mole._NTR = resnumb

    #chains = copy(chains_res['A'])
    #for res in chains:
    #    if str(res) > config.terminal_offset:
    #        print((res, chains_res['A'][res]))
    #        del chains_res['A'][res]
    #chains_res['A'][config.tit_mole._NTR] = 'NTR'
    #chains_res['A'][config.tit_mole._CTR] = 'CTR'

    with open('cleaned.pqr', 'w') as f_new:
        f_new.write(new_pdb_text)

    with open('removed.pqr', 'w') as f_new:
        f_new.write(removed_pdb_text)

    sites_addHtaut = ''
    for chain in molecules:
        for res in chains_res[chain]:
            if chains_res[chain][res] == 'NTR' or \
               (chain in CYS_bridges and res in CYS_bridges[chain]):
               continue
            sites_addHtaut += '{0}-{1}-{2},'.format(res, chains_res[chain][res],
                                                    chain.replace(' ', '_'))

    if len(sites_addHtaut) > 0 and sites_addHtaut[-1] == ',':
        sites_addHtaut = sites_addHtaut[:-1]

    logfile = 'LOG_addHtaut'
    #log.redirectErr("start", logfile)

    script_dir = Config.pypka_params['script_dir']
    os.system('{}/addHtaut_{} cleaned.pqr {} > {} 2> {}'.format(script_dir,
                                                                Config.pypka_params['ff_family'],
                                                                sites_addHtaut,
                                                                outputpqr, logfile))

    #log.redirectErr("stop", logfile)
    log.checkDelPhiErrors(logfile, 'addHtaut')

    with open(outputpqr) as f:
        content = f.read()
    with open(outputpqr, 'w') as f_new:
        f_new.write(content + removed_pdb_text)

    with open(pdb_filename) as f:
        box = None
        for line in f:
            if line.startswith('CRYST1'):
                tmp = line.split()[1:4]
                box = (float(tmp[0]), float(tmp[1]), float(tmp[2]))
        if not box:
            box = (1.0, 1.0, 1.0)

    #pdb2gro(outputpqr, "TMP.gro", chains_res, box, pqr=True, fix_termini=False)

    with open(outputpqr) as f,\
         open('TMP.pdb', 'w') as fnew:
        counter = 0
        new_pdb_text = "CRYST1  {:3.3f}  {:3.3f}  {:3.3f}  "\
                       "60.00  60.00  90.00 P 1           1\n".format(box[0],
                                                                      box[1], box[2])
        for line in f:
            counter += 1
            (aname, anumb_old, resname, chain, resnumb, x, y,
             z, charge, radius) = read_pqr_line(line)

            if chain == '_':
                chain = ' '

            new_pdb_text += new_pdb_line(counter, aname, resname, resnumb, x, y, z, chain=chain)
        fnew.write(new_pdb_text)

    keep_membrane = False
    keep_ions = False
    if Config.delphi_params['pbc_dim'] == 2:
        keep_membrane = True

    if Config.pypka_params['keep_ions']:
        keep_ions = True
    if keep_ions or keep_membrane:
        add_non_protein(pdb_filename, 'TMP.pdb',
                        keep_membrane=keep_membrane, keep_ions=keep_ions)

    tmpfiles = ('LOG_pdb2pqr', 'LOG_pdb2pqr_err', 'clean.pqr', 'cleaned.pqr',
                'cleaned_tau.pqr', 'input_clean.pdb', 'removed.pqr', 'Hs.pqr')
    for filename in tmpfiles:
        if not Config.debug and os.path.isfile(filename):
            os.remove(filename)

def removeMembrane(pdbfile):
    nomembrane_text = ''
    with open(pdbfile) as f:
        for line in f:
            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)

                if chain == ' ':
                    chain = '_' # workaround to deal with pdb2pqr

                if resname not in LIPID_RESIDUES:
                    nomembrane_text += new_pdb_line(anumb, aname, resname, resnumb,
                                                    x, y, z, chain=chain)

            else:
                nomembrane_text += line
    with open('tmp.tmp', 'w') as f_new:
        f_new.write(nomembrane_text)
    os.rename('tmp.tmp', 'input_clean.pdb')
    return 'input_clean.pdb'


def add_non_protein(pdbfile_origin, add_to_pdb, keep_membrane=False, keep_ions=False):
    new_file_body = ''

    with open(add_to_pdb) as f:
        for line in f:
            if line.startswith('ATOM '):
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)
                last_anumb = anumb
                last_resnumb = resnumb

    # Read the original pdb with the membrane
    with open(pdbfile_origin) as f:
        for line in f:
            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)
                if keep_membrane:
                    if resname in LIPID_RESIDUES:
                        last_anumb += 1
                        new_file_body += new_pdb_line(last_anumb, aname, resname,
                                                      resnumb, x, y, z, chain=chain)

                    if resname in list(Config.pypka_params.LIPIDS.values()):
                        aname, resname, to_include = convert_FF_atomnames(aname, resname)
                        if to_include:
                            last_anumb += 1
                            resnumb += last_resnumb
                            new_file_body += new_pdb_line(last_anumb, aname, resname,
                                                          resnumb, x, y, z, chain=chain)
                if keep_ions and aname in IONS and resname == aname:
                    last_anumb += 1
                    resnumb += last_resnumb
                    new_file_body += new_pdb_line(last_anumb, aname, resname,
                                                  resnumb, x, y, z, chain=chain)

    with open(add_to_pdb, 'a') as f_new:
        f_new.write(new_file_body)

def convert_FF_atomnames(aname, resname):
    popc_resname = Config.pypka_params.LIPIDS['POPC']
    chol_resname = Config.pypka_params.LIPIDS['cholesterol']

    lookup = [((' N  ', popc_resname), ('NTM', 'CHL')),
              ((' C12', popc_resname), ('CB ', 'CHL')),
              ((' C13', popc_resname), ('CN1', 'CHL')),
              ((' C14', popc_resname), ('CN2', 'CHL')),
              ((' C15', popc_resname), ('CN3', 'CHL')),
              ((' C11', popc_resname), ('CA ', 'PJ2')),
              ((' P  ', popc_resname), ('P  ', 'PJ2')),
              ((' O13', popc_resname), ('OG ', 'PJ2')),
              ((' O14', popc_resname), ('OB ', 'PJ2')),
              ((' O12', popc_resname), ('OA ', 'PJ2')),
              ((' O11', popc_resname), ('OD ', 'PJ2')),
              ((' C1 ', popc_resname), ('CD ', 'POX')),
              ((' C2 ', popc_resname), ('CE ', 'POX')),
              ((' O21', popc_resname), ('OE ', 'POX')),
              ((' C21', popc_resname), ('C1A', 'POX')),
              ((' O22', popc_resname), ('O1A', 'POX')),
              ((' C22', popc_resname), ('C1B', 'POX')),
              ((' C3 ', popc_resname), ('CZ ', 'POX')),
              ((' O31', popc_resname), ('OZ ', 'POX')),
              ((' C31', popc_resname), ('C2A', 'POX')),
              ((' O32', popc_resname), ('O2A', 'POX')),
              ((' C32', popc_resname), ('C2B', 'POX')),
              ((' C23', popc_resname), ('C1C', 'POX')),
              ((' C24', popc_resname), ('C1D', 'POX')),
              ((' C25', popc_resname), ('C1E', 'POX')),
              ((' C26', popc_resname), ('C1F', 'POX')),
              ((' C27', popc_resname), ('C1G', 'POX')),
              ((' C28', popc_resname), ('C1H', 'POX')),
              ((' C29', popc_resname), ('C1I', 'POX')),
              (('0C21', popc_resname), ('C1J', 'POX')),
              (('1C21', popc_resname), ('C1K', 'POX')),
              (('2C21', popc_resname), ('C1L', 'POX')),
              (('3C21', popc_resname), ('C1M', 'POX')),
              (('4C21', popc_resname), ('C1N', 'POX')),
              (('5C21', popc_resname), ('C1O', 'POX')),
              (('6C21', popc_resname), ('C1P', 'POX')),
              (('7C21', popc_resname), ('C1Q', 'POX')),
              (('8C21', popc_resname), ('C1R', 'POX')),
              ((' C33', popc_resname), ('C2C', 'POX')),
              ((' C34', popc_resname), ('C2D', 'POX')),
              ((' C35', popc_resname), ('C2E', 'POX')),
              ((' C36', popc_resname), ('C2F', 'POX')),
              ((' C37', popc_resname), ('C2G', 'POX')),
              ((' C38', popc_resname), ('C2H', 'POX')),
              ((' C39', popc_resname), ('C2I', 'POX')),
              (('0C31', popc_resname), ('C2J', 'POX')),
              (('1C31', popc_resname), ('C2K', 'POX')),
              (('2C31', popc_resname), ('C2L', 'POX')),
              (('3C31', popc_resname), ('C2M', 'POX')),
              (('4C31', popc_resname), ('C2N', 'POX')),
              (('5C31', popc_resname), ('C2O', 'POX')),
              (('6C31', popc_resname), ('C2P', 'POX')),
              ((" H3", chol_resname), (" H3", 'CHO')),
              ((" C3", chol_resname), (" C3", 'CHO')),
              ((" C2", chol_resname), (" C2", 'CHO')),
              ((" C1", chol_resname), (" C1", 'CHO')),
              (("C10", chol_resname), ("C10", 'CHO')),
              (("C19", chol_resname), ("C19", 'CHO')),
              ((" C9", chol_resname), (" C9", 'CHO')),
              (("C11", chol_resname), ("C11", 'CHO')),
              (("C12", chol_resname), ("C12", 'CHO')),
              (("C13", chol_resname), ("C13", 'CHO')),
              (("C18", chol_resname), ("C18", 'CHO')),
              (("C17", chol_resname), ("C17", 'CHO')),
              (("C20", chol_resname), ("C20", 'CHO')),
              (("C22", chol_resname), ("C22", 'CHO')),
              (("C23", chol_resname), ("C23", 'CHO')),
              (("C24", chol_resname), ("C24", 'CHO')),
              (("C25", chol_resname), ("C25", 'CHO')),
              (("C27", chol_resname), ("C27", 'CHO')),
              (("C26", chol_resname), ("C26", 'CHO')),
              (("C21", chol_resname), ("C21", 'CHO')),
              (("C16", chol_resname), ("C16", 'CHO')),
              (("C15", chol_resname), ("C15", 'CHO')),
              (("C14", chol_resname), ("C14", 'CHO')),
              ((" C8", chol_resname), (" C8", 'CHO')),
              ((" C7", chol_resname), (" C7", 'CHO')),
              ((" C6", chol_resname), (" C6", 'CHO')),
              ((" H6", chol_resname), (" H6", 'CHO')),
              ((" C5", chol_resname), (" C5", 'CHO')),
              ((" C4", chol_resname), (" C4", 'CHO')),
              ((" O3", chol_resname), (" O3", 'CHO')),
              (("H3'", chol_resname), ("H3'", 'CHO'))]

    for swap in lookup:
        old_aname = swap[0][0].strip()
        old_resname = swap[0][1]
        new_aname = swap[1][0].strip()
        new_resname = swap[1][1]
        if old_aname == aname and \
           old_resname == resname:
            aname = new_aname
            resname = new_resname
            return aname, resname, True

    return aname, resname, False



from config import Config
from constants import TERMINAL_OFFSET, TITRABLETAUTOMERS, PROTEIN_RESIDUES

def new_pdb_line(aID, aname, resname, resnumb, x, y, z, chain=' '):
    pdb_format = "ATOM  {:5d} {:4s} {:4s}{:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}\n"
    return pdb_format.format(aID, aname, resname, chain, resnumb, x, y, z)

def new_pqr_line(aID, aname, resname, resnumb, x, y, z, charge, radius, chain=' '):
    pdb_format = "ATOM  {:5d} {:4s} {:4s}{:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}\n"
    return pdb_format.format(aID, aname, resname, chain, resnumb, x, y, z, charge, radius)

def new_gro_line(aID, aname, resname, resnumb, x, y, z):
    gro_format = '{:5}{:5}{:>5}{:5}{:8.3f}{:8.3f}{:8.3f}\n'
    return gro_format.format(resnumb, resname, aname, aID, x, y, z)


def correct_longHs(aname):
    if aname[1] == 'H' and aname[0] in '12' \
       and aname[3] in '12':
        return aname[1:3] + aname[3] + aname[0]
    return aname


def read_pqr_line(line):
    aname   = line[12:16].strip()
    anumb   = int(line[7:11].strip())
    resname = line[17:21].strip()
    chain   = line[21]
    resnumb = int(line[22:26])
    x       = float(line[30:38])
    y       = float(line[38:46])
    z       = float(line[46:54])
    charge  = float(line[54:62])
    radius  = float(line[62:70])
    if len(aname) == 4:
        aname = correct_longHs(aname)
    return (aname, anumb, resname, chain, resnumb, x, y, z, charge,
            radius)

def read_pdb_line(line):
    aname   = line[12:16].strip()
    anumb   = int(line[5:11].strip())
    resname = line[17:21].strip()
    chain   = line[21]
    resnumb = int(line[22:26])
    x       = float(line[30:38])
    y       = float(line[38:46])
    z       = float(line[46:54])
    if len(aname) == 4:
        aname = correct_longHs(aname)
    return (aname, anumb, resname, chain, resnumb, x, y, z)

def read_gro_line(line):
    resnumb = int(line[:5].strip())
    resname = line[5:10].strip()
    aname   = line[10:15].strip()
    anumb   = int(line[15:20].strip())
    x = float(line[20:28].strip())
    y = float(line[28:36].strip())
    z = float(line[36:44].strip())
    return (aname, anumb, resname, resnumb, x, y, z)

def gro2pdb(f_in, f_out, save_box=False):
    new_pdb_content = "REMARK    CONVERTED from {} by PypKa\n".format(f_in)
    with open(f_in) as f:
        line_counter = 0
        natoms_left = 0
        natoms = 0
        lines = f.read().splitlines()
        penul_line = lines[-2]
        last_line  = lines[-1]
        box = [float(i) * 10 for i in last_line.split()[:3]]
        if save_box:
            Config.pypka_params.setBox(box)
        new_pdb_content += "CRYST1  {0:3.3f}  {1:3.3f}  {2:3.3f}  "\
                           "60.00  60.00  90.00 P 1           1\n".format(box[0],
                                                                          box[1], box[2])
        for line in lines:
            line_counter += 1
            if natoms_left > 0:
                natoms_left -= 1
                natoms += 1
                (aname, anumb, resname,
                 resnumb, x, y, z) = read_gro_line(line)

                x, y, z = x * 10, y * 10, z * 10

                if resname[-1] == 'X' and Config.pypka_params.CpHMD_mode:
                    resname = resname[:-1] + '0'

                new_pdb_content += new_pdb_line(anumb, aname,
                                                resname, resnumb,
                                                x, y, z, chain='A')
            elif line_counter == 2:
                natoms_left = int(line.strip())
    new_pdb_content += 'TER\nENDMDL\n'
    with open(f_out, 'w') as f_new:
        f_new.write(new_pdb_content)

def pdb2gro(filename_in, filename_out, chains_res, box=[], pqr=False,
            renumber_res=False, fix_termini=True):
    """
    Returns
      - aposition (int) of last atom id in filename_out
    """
    NTR_atoms = Config.pypka_params['NTR_atoms']
    CTR_atoms = Config.pypka_params['CTR_atoms']

    header = 'CREATED within PyPka\n'
    new_pdb_text = ''
    aposition = 0
    rposition = 1
    prev_resnumb = None

    with open(filename_in) as f:
        for line in f:
            if 'CRYST1' in line[:6]:
                parts = line.split()
                pdb_box = [float(parts[1]), float(parts[2]), float(parts[3])]
                continue
            elif 'ATOM' != line[:4]:
                continue
            elif pqr:
                (aname, anumb, resname, chain, resnumb, x, y,
                 z, charge, radius) = read_pqr_line(line)
            else:
                (aname, anumb, resname, chain,
                 resnumb, x, y, z) = read_pdb_line(line)
            aposition += 1

            chains_sites_numbs = []
            if chain in chains_res:
                chains_sites_numbs = chains_res[chain]
                for site_numb, site_name in chains_sites_numbs.items():
                    if site_name == 'NTR':
                        NTR_numb = int(site_numb)
                    elif site_name == 'CTR':
                        CTR_numb = int(site_numb)

            if fix_termini and resnumb in chains_sites_numbs:
                if (resnumb == NTR_numb and aname in NTR_atoms) or \
                   (resnumb == CTR_numb and aname in CTR_atoms):
                    site_numb = resnumb
                    resname = chains_res[chain][site_numb]

            elif (resname == 'HIS' and aname == 'HD1' and
                  resnumb not in chains_sites_numbs):
                aposition -= 1
                continue

            if prev_resnumb and prev_resnumb != resnumb:
                rposition += 1

            x /= 10.0
            y /= 10.0
            z /= 10.0
            if not renumber_res:
                rposition = resnumb
            new_pdb_text += new_gro_line(aposition, aname, resname,
                                         rposition, x, y, z)
            prev_resnumb = resnumb

    header += '{0}\n'.format(aposition)
    if box == []:
        footer = '{0:10.5f}{1:10.5f}{2:10.5f}\n'.format(pdb_box[0] / 10.0,
                                                        pdb_box[1] / 10.0,
                                                        pdb_box[2] / 10.0)
    else:
        footer = '{0:10.5f}{1:10.5f}{2:10.5f}\n'.format(box[0] / 10.0,
                                                        box[1] / 10.0,
                                                        box[2] / 10.0)

    new_pdb = header + new_pdb_text + footer
    with open(filename_out, 'w') as f_new:
        f_new.write(new_pdb)

    return aposition


def correct_names(resnumb, resname, aname,
                  titrating_sites, NTR_numb, CTR_numb):
    def change_aname(aname, restype, mode='regular'):
        if mode == 'titrating':
            not_correct_names = list(correct_atoms_sites_table[restype].keys())
        else:
            not_correct_names = list(correct_atoms_table[restype].keys())
        for not_corrected in not_correct_names:
            if aname == not_corrected:
                if mode == 'titrating':
                    aname = correct_atoms_sites_table[restype][not_corrected]
                else:
                    aname = correct_atoms_table[restype][not_corrected]

        return aname

    def change_resname(resname):
        for tit_res in TITRABLETAUTOMERS.keys():
            if tit_res[:2] == resname[:2]:
                return tit_res
        return resname


    # no longer used as it is done by pdb2pqr
    correct_atoms_table = {'CTR': {'O': 'O1',
                                   'OXT': 'O2'},
                           'NTR': {'H': 'H1'},
                           'ILE': {'CD1': 'CD'}}

    correct_atoms_sites_table = {'CYS': {'HG': 'HG1'},
                                 'SER': {'HG': 'HG1'},
                                 'TYR': {'HH': 'HH1'}}

    correct_residues_table = {'HSD': 'HI0',
                              'HSE': 'HI1',
                              'HSP': 'HI2',
                              'ARGN': 'AR0',
                              'ASPH': 'AS0',
                              'CYS2': 'CYS',
                              'CYSH': 'CY0',
                              'GLUH': 'GL0',
                              'HISD': 'HI0',
                              'HISE': 'HI1',
                              'HISH': 'HI2',
                              'HISA': 'HI0',
                              'HISB': 'HI1',
                              'HISP': 'HI2',
                              'LYSH': 'LY3',
                              'LYSN': 'LY0',
                              'ISU': 'CYS'}

    if resnumb == CTR_numb:
        restype = 'CTR'
        aname = change_aname(aname, restype)

    if resnumb == NTR_numb:
        restype = 'NTR'
        aname = change_aname(aname, restype)

    if resname in list(correct_atoms_table.keys()):
        aname = change_aname(aname, resname)

    if resname in list(correct_residues_table.keys()):
        resname = correct_residues_table[resname]

    if resnumb in titrating_sites:
        if resname not in PROTEIN_RESIDUES:
            resname = change_resname(resname)
        if resname in list(correct_atoms_sites_table.keys()):
            aname = change_aname(aname, resname, mode='titrating')


    return aname, resname


def readPBPFile(f_dat):
    with open(f_dat) as f:
        for line in f:
            if line[0] != "#" and len(line) > 1:
                parts = line.strip().split('=')
                param = parts[0].split()[-1]
                value = parts[1].replace('"', '').replace("'", '')
                config.params[param] = value.strip()


def convertTermini(site_numb):
    if site_numb >= TERMINAL_OFFSET:
        return site_numb - TERMINAL_OFFSET
    return site_numb



"""Identifying titrable sites and integrity checking functions
"""
from config import Config
from constants import *
from formats import read_pdb_line, read_gro_line, new_pdb_line
from ffconverter import main_chains
from copy import copy

def identify_tit_sites(molecules, instanciate_sites=True):
    def SitesFileLine(resnumb, resname):
        for res in REGULARTITRATINGRES:
            if res[0:2] == resname[0:2]:
                resname = res
        if resname in TITRABLETAUTOMERS:
            ntautomers = TITRABLETAUTOMERS[resname]
        else:
            for res in REGULARTITRATINGRES:
                if res[0:2] == resname[0:2]:
                    ntautomers = TITRABLETAUTOMERS[res]
        # In debug mode, a .sites file is created
        text = '{0} '.format(resnumb)
        for i in range(ntautomers):
            text += '{0}tau{1} '.format(resname, i + 1)

        if res in ('NTR', 'CTR'):
            resnumb += TERMINAL_OFFSET

        if instanciate_sites:
            instanciate_site(resnumb, resname, ntautomers)
        return text + '\n'

    def instanciate_site(resnumb, resname, ntautomers):
        sID = molecule.addSite(resnumb)
        molecule.addTautomers(sID, ntautomers, resname)

    def add2chain(chain, chain_res, resnumb, resname):
        chain_res[chain][resnumb] = resname

    sites_file = ''
    sites = {chain: [] for chain in molecules.keys()}
    chain_res = {chain: {} for chain in molecules.keys()}
    last_res = None
    with open(Config.pypka_params['f_in']) as f:
        nline = 0
        resnumb = None
        resname = None
        for line in f:
            nline += 1
            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname, chain,
                 resnumb, x, y, z) = read_pdb_line(line)

                if line[26] != ' ':
                    continue
                #if chain == ' ':
                #    chain = 'A'
                last_res = resnumb
                last_chain = chain
                chain_sites = []

                if chain in molecules.keys():
                    molecule = molecules[chain]
                    chain_sites = chain_res[chain]

                    if resname in PROTEIN_RESIDUES or \
                       resname in TITRABLERESIDUES:
                        if resnumb not in chain_sites:
                            if not chain_res[chain]:
                                sites_file += SitesFileLine(resnumb, 'NTR')
                                add2chain(chain, chain_res, str(resnumb), 'NTR')
                                if instanciate_sites:
                                    molecule.NTR = resnumb

                            if resname in TITRABLERESIDUES and \
                               resname != 'NTR' and resname != 'CTR':
                                if Config.pypka_params['ser_thr_titration'] == False and \
                                   resname in ('SER', 'THR'):
                                    continue
                                sites_file += SitesFileLine(resnumb, resname)
                                add2chain(chain, chain_res, resnumb, resname)

                        if 'CTR' not in sites and \
                           aname in ('CT', 'OT', 'OT1', 'OT2', 'O1', 'O2', 'OXT'):
                            sites_file += SitesFileLine(resnumb, 'CTR')
                            add2chain(chain, chain_res, str(resnumb), 'CTR')
                            if instanciate_sites:
                                molecule.CTR = resnumb


    if 'CTR' not in sites and \
       aname in ('CT', 'OT', 'OT1', 'OT2', 'O1', 'O2', 'OXT') and molecule.sites == 'all':
        sites_file += SitesFileLine(last_res, 'CTR')
        add2chain(last_chain, chain_res, str(last_res), 'CTR')
        if instanciate_sites:
            molecule = molecules[chain]
            molecule.CTR = resnumb

    if not chain_res and instanciate_sites:
        f_in = Config.pypka_params['f_in']
        raise Exception('Not one titrable residue was found in {}'.format(f_in))

    if instanciate_sites:
        for molecule in molecules.values():
            # Adding the reference tautomer to each site
            molecule.addReferenceTautomers()
            # Assigning a charge set to each tautomer
            molecule.addTautomersChargeSets()

    if Config.debug:
        with open('tmp.sites', 'w') as f_new:
            f_new.write(sites_file)

    return chain_res


def check_sites_integrity(molecules, chains_res, useTMPpdb=False):
    """Identifies titrable residues and checks integrity of the residue blocks
    (excluding Hydrogens)
    """
    def check_site(prev_resname, cur_atoms, ter=None):
        def correctResName(resname):
            for res in REGULARTITRATINGRES:
                if res[0:2] == resname[0:2]:
                    return res
            return resname

        def makeSite(molecule, resnumb, resname, termini_resname=None):
            if resname in TITRABLETAUTOMERS:
                ntautomers = TITRABLETAUTOMERS[resname]
            else:
                for res in REGULARTITRATINGRES:
                    if res[0:2] == resname[0:2]:
                        ntautomers = TITRABLETAUTOMERS[res]
            sID = molecule.addSite(resnumb)
            molecule.addTautomers(sID, ntautomers, resname,
                                  termini_resname=termini_resname)
            #print('added', molecule.chain, resnumb, resname)

        prev_resname = correctResName(prev_resname)

        res_tits = True
        if ter:
            if (not Config.pypka_params['ser_thr_titration'] and 
                prev_resname in ('SER', 'THR')):
                res_tits = False
            else:
                res_tits = bool(prev_resname in TITRABLERESIDUES)

        res_atoms = copy(cur_atoms)
        (integrity_terminal,
         integrity_site) = check_integrity(prev_resname,
                                           res_atoms,
                                           ter=ter,
                                           site=res_tits)


        if integrity_terminal:
            ter_resnumb = prev_resnumb + TERMINAL_OFFSET
            makeSite(molecule, ter_resnumb, ter, termini_resname=prev_resname)
            if ter == 'NTR':
                molecule.NTR = prev_resnumb
            elif ter == 'CTR':
                molecule.CTR = prev_resnumb
        else:
            warning(molecule, prev_resnumb, ter, '')

        if prev_resnumb in sites:
            if integrity_site:
                makeSite(molecule, prev_resnumb, prev_resname)
            else:
                warning(molecule, prev_resnumb,
                        prev_resname, cur_atoms)
        elif prev_resname == 'CYS': # dealing with a CYS that is not in sites
            if not integrity_site:
                warning(molecule, prev_resnumb, prev_resname, cur_atoms, mode='CYS')


    def warning(molecule, resnumb, resname, res_atoms, mode=None):
        if mode == 'CYS' or resname == 'CYS':
            CYS_atoms = ['N', 'CA', 'CB', 'SG', 'C', 'O', 'H']
            if set(res_atoms).issubset(CYS_atoms) and \
               set(CYS_atoms).issubset(res_atoms):
                # no need to correct residue name
                warn = '{0} {1} is assumed to be participating '\
                       'in a SS-bond'.format(resnumb, resname)
                Config.log.report2log(warn)
                return
            CY0_atoms = ['N', 'CA', 'CB', 'SG', 'C', 'O', 'H', 'HG1']
            if set(res_atoms).issubset(CY0_atoms) and \
               set(CY0_atoms).issubset(res_atoms):
                molecule.correct_names[resnumb] = 'CY0'
                return
            CY0_atoms = ['N', 'CA', 'CB', 'SG', 'C', 'O', 'H', 'HG']
            if set(res_atoms).issubset(CY0_atoms) and \
               set(CY0_atoms).issubset(res_atoms):
                molecule.correct_names[resnumb] = 'CY0'
                molecule.correct_atoms[resnumb] = {'HG': 'HG1'}
                return
            else:
                warn = '{0} {1} failed integrity check'.format(resnumb,
                                                               resname)
                Config.log.report2log(warn)
        elif resname not in TITRABLERESIDUES:
            return
        else:
            warn = '{0} {1} failed integrity check'.format(resnumb,
                                                           resname)
            Config.log.report2log(warn)


    if Config.pypka_params['f_in'] and not useTMPpdb:
        filename = Config.pypka_params['f_in']
        filetype = 'pdb'
    else:
        filename = "TMP.pdb"
        filetype = 'pdb'

    resnumb = None
    cur_atoms = []
    prev_resnumb = None
    prev_resname = None
    last_chain = None
    with open(filename) as f:
        nline = 0
        f_lines = f.readlines()
        maxnlines = len(f_lines)
        for line in f_lines:
            resname = None
            nline += 1

            chain = None

            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname, chain,
                 resnumb, x, y, z) = read_pdb_line(line)

                if chain in molecules:
                    if not last_chain:
                        last_chain = chain
                    molecule = molecules[last_chain]
                    sites = chains_res[last_chain]
                    last_molecule = molecule
                    last_chain = chain

                if nline == maxnlines:
                    cur_atoms.append(aname)

            if line == 'TER\n':
                resnumb += 1

            if (prev_resnumb != resnumb or nline == maxnlines) and \
               prev_resnumb is not None:
                if nline == maxnlines:
                    prev_resnumb = copy(resnumb)
                    resnumb = 'None'

                if chain in molecules:
                    if prev_resname in TITRABLERESIDUES or \
                       (prev_resnumb == molecule.NTR or resnumb == molecule.NTR) or \
                       (prev_resnumb == molecule.CTR or resnumb == molecule.CTR):
                        if prev_resnumb == molecule.NTR and resnumb != molecule.NTR:
                            check_site(prev_resname, cur_atoms, ter='NTR')
                            prev_resnumb = None
                        # Dealing with the last residue and CTR
                        elif prev_resnumb == molecule.CTR and resnumb != molecule.CTR:
                            check_site(prev_resname, cur_atoms, ter='CTR')
                        # Dealing with the previous residue
                        elif prev_resnumb is not None and \
                             prev_resname in TITRABLERESIDUES:
                            if not (not Config.pypka_params['ser_thr_titration'] and \
                                prev_resname in ('SER', 'THR')):
                                check_site(prev_resname, cur_atoms)

                    elif prev_resname == 'ALA':
                        # TODO: check residue block integrity for other non titrating residues
                        pass
                elif last_molecule and prev_resnumb == last_molecule.CTR and resnumb != last_molecule.CTR:
                    check_site(prev_resname, cur_atoms, ter='CTR')
                elif prev_resname in TITRABLERESIDUES and prev_resnumb is not None:
                    check_site(prev_resname, cur_atoms)

            # Dealing with the new residue
            if prev_resnumb != resnumb:
                cur_atoms = [aname]
                prev_resnumb = resnumb
                prev_resname = resname
            elif resnumb is not None:
                cur_atoms.append(aname)
                if prev_resname in ('NTR', 'CTR') and \
                   prev_resname != resname:
                    prev_resname = resname

    for molecule in molecules.values():
        # Adding the reference tautomer to each site
        molecule.addReferenceTautomers()
        # Assigning a charge set to each tautomer
        molecule.addTautomersChargeSets()

    # TODO: report blocks that failed the check (in .log file with
    # numbering reference to stepwise scheme)
    # TODO: add lipid residues
    if Config.debug:
        print('exiting check_sites_integrity')

def check_integrity(resname, res_atoms,
                    ter=False, site=True):
    def read_anames(filename):
        res_atoms_st = []
        with open(filename) as f:
            for line in f.readlines()[1:]:
                cols = line.strip().split()
                aname = cols[1]
                res_atoms_st.append(aname)
        return res_atoms_st
    def pop_atoms(anames1, anames2):
        trigger = False
        for aname in anames1:
            if aname not in anames2:
                trigger = True
                if Config.debug:
                    print((aname, 'not in', resname))
            else:
                anames2.remove(aname)
        if trigger:
            return anames2, False
        else:
            return anames2, True
    if Config.debug:
        print('###### INTEGRITY CHECK ######')
        print(res_atoms)
    integrity = None
    integrity_ter = None

    if ter:
        st = '{0}/{1}/sts/{2}tau1.st'.format(Config.pypka_params['script_dir'],
                                             Config.pypka_params['ffID'],
                                             ter)
        res_atoms_st = read_anames(st)
        res_atoms, integrity_ter = pop_atoms(res_atoms_st, res_atoms)

    if site:
        main_chain_atoms = main_chains[Config.pypka_params['ff_family']]
        if ter == 'NTR':
            main_chain = main_chain_atoms['NTR']
        elif ter == 'CTR':
            main_chain = main_chain_atoms['CTR']
        else:
            main_chain = main_chain_atoms['main']

        for aname in main_chain:
            if aname in res_atoms:
                res_atoms.remove(aname)
            elif not ter:
                integrity = False

        if Config.debug:
            print(('i', integrity))
        if integrity is not False:
            filename = '{0}/{1}/sts/{2}tau1.st'.format(Config.pypka_params['script_dir'],
                                                       Config.pypka_params['ffID'],
                                                       resname)
            res_atoms_st = read_anames(filename)
            res_atoms, integrity = pop_atoms(res_atoms_st, res_atoms)
            if Config.debug:
                print((res_atoms_st, res_atoms))
                print(('i', integrity))

    if len(res_atoms) != 0 and integrity:
        raise Exception('Something is wrong')

    return integrity_ter, integrity

def make_delphi_inputfile(f_in, f_out, molecules):
    def getMaxCoords(coords, max_coords):
        x, y, z = coords
        max_x, max_y, max_z = max_coords
        if max_x < x:
            max_x = x
        if max_y < y:
            max_y = y
        if max_z < z:
            max_z = z
        return max_x, max_y, max_z

    def correct_termini(resnumb, resname, aname, ntr_res, ctr_res):
        if resnumb == ntr_res and \
           aname in Config.pypka_params['NTR_atoms']:
            resname = 'NTR'
            resnumb += TERMINAL_OFFSET
        elif resnumb == ctr_res and \
             aname in Config.pypka_params['CTR_atoms']:
            resname = 'CTR'
            resnumb += TERMINAL_OFFSET
            if aname == 'C':
                aname = 'CT'
        return resnumb, resname, aname

    def correct_res_names(molecule, resnumb, resname, aname):
        if resnumb in list(molecule.correct_names.keys()):
            resname = molecule.correct_names[resnumb]
        if resnumb in list(molecule.correct_atoms.keys()) and \
           aname in molecule.correct_atoms[resnumb]:
            aname = molecule.correct_atoms[resnumb][aname]

        return resnumb, resname, aname

    def assign_atoms(sites, resnumb, aname, site_Hs, site_positions):
        ref_tau_name = resname
        if resnumb in list(sites.keys()) and \
           aname in list(sites[resnumb].getRefTautomer().charge_set.keys()):
            #( aname not in ('N', 'H', 'C', 'O', 'CA') or
            #(aname in ('N', 'H', 'C', 'O', 'CA') and resname == 'NTR')):
            # change res name to reference tautomer
            ref_tau_name = sites[resnumb].getRefTautomerName()

            # add atom to corresponding site
            sites[resnumb].addAtom(aname, anumb)

            if chain not in site_positions:
                site_positions[chain] = {}
                site_Hs[chain] = {}
            if resnumb not in site_positions[chain]:
                site_positions[chain][resnumb] = []
                site_Hs[chain][resnumb] = []

            if resnumb in site_positions[chain]:
                site_positions[chain][resnumb].append((x, y, z))
                if aname[0] == 'H':
                    site_Hs[chain][resnumb].append((x, y, z))

        return site_Hs, ref_tau_name, site_positions

    new_pdb_content = ""
    site_positions = {}
    site_Hs = {}
    max_box = [0.0, 0.0, 0.0]
    aposition = -1
    with open(f_in) as f:
        for line in f:
            if line.startswith('ATOM'):
                aposition += 1
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)

                max_box = getMaxCoords([x, y, z], max_box)

                if chain in molecules:
                    molecule = molecules[chain]
                    ntr_res = molecule.NTR
                    ctr_res = molecule.CTR
                    sites = molecule.sites

                    if (resname == 'HIS' and aname == 'HD1' and
                        resnumb not in sites.keys()):
                        aposition -= 1
                        continue

                    resnumb, resname, aname = correct_termini(resnumb, resname, aname,
                                                              ntr_res, ctr_res)

                    resnumb, resname, aname = correct_res_names(molecule, resnumb,
                                                                resname, aname)

                    titrable_res = False
                    if resnumb in sites.keys():
                        titrable_res = True

                    molecule.addAtom(aname, anumb, aposition, titrable_res)

                    (site_Hs, resname,
                     site_positions) = assign_atoms(sites, resnumb, aname,
                                                    site_Hs, site_positions)

                else:
                    if (resname == 'HIS' and aname == 'HD1'):
                        aposition -= 1
                        continue
                    resnumb, resname, aname = correct_res_names(molecule, resnumb,
                                                                resname, aname)

                new_pdb_content += new_pdb_line(aposition, aname,
                                                resname, resnumb,
                                                x, y, z)

            elif line.startswith('CRYST1'):
                parts = line.split()
                box = [float(i) for i in parts[1:4]]


    if box == [0.1, 0.1, 0.1]:
        box = max_box

    if Config.pypka_params['box']:
        box = Config.pypka_params['box']
    else:
        Config.pypka_params.setBox(box)


    if Config.delphi_params['pbc_dim'] == 2:
        Config.delphi_params.redefineScale()

    new_pdb_content += 'TER\nENDMDL\n'
    with open(f_out, 'w') as f_new:
        f_new.write(new_pdb_content)

    # TODO: check Terminal_offset has to be bigger than the total number of residues
    # TODO: delete terminal_offset and use another approach to distinguish between N- and C-ter
    # TODO: check size xy > config.cutoff * 2
    # if so, raise Exception, and ask to change cutoff value

    # TODO: check if pbc_dim -> set gsizes from pdb size xy and ignore perfil

    for chain in site_positions.keys():
        molecule = molecules[chain]
        for site in site_positions[chain]:
            if site in list(molecule.sites.keys()):
                pos_max = [-9999990, -999999, -999999]
                pos_min = [999999, 999999, 999999]
                focus_center = [0, 0, 0]
                for atom in site_positions[chain][site]:
                    for i in range(3):
                        if pos_max[i] < atom[i]:
                            pos_max[i] = atom[i]
                        if pos_min[i] > atom[i]:
                            pos_min[i] = atom[i]
                focus_center[0] = (pos_max[0] + pos_min[0]) / 2
                focus_center[1] = (pos_max[1] + pos_min[1]) / 2
                focus_center[2] = (pos_max[2] + pos_min[2]) / 2

                if Config.delphi_params['pbc_dim'] == 2:
                    molecule.sites[site].addCenter(focus_center,
                                                   boxsize=box[0],
                                                   box_z=box[2])
                else:
                    molecule.sites[site].addCenter(focus_center)
                hx, hy, hz = 0, 0, 0
                nHs = len(site_Hs[chain][site])
                if nHs == 0:
                    sitename = molecule.sites[site].getName()
                    raise Exception('Site {1}{0} appears '
                                    'to have no Hydrogen atoms'.format(site,
                                                                           sitename))
                for h in site_Hs[chain][site]:
                    hx += h[0]
                    hy += h[1]
                    hz += h[2]
                hx /= nHs
                hy /= nHs
                hz /= nHs
                Hcenter = (hx, hy, hz)
                molecule.sites[site].addCenterH(Hcenter)


def get_chains_from_file(f_in):
    chain_list = []
    with open(f_in) as f:
        for line in f:
            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname, chain,
                 resnumb, x, y, z) = read_pdb_line(line)
                if chain not in chain_list:
                    chain_list.append(chain)

    return chain_list



import os
import sys
from config import Config

class Log:
    def __init__(self):
        self.f_log       = "LOG"
        self.stdout      = None
        self.stderr      = None
        self.stdout_file = None
        self.stderr_file = None

    def report2log(self, info, stdout=False):
        with open(self.f_log, 'a') as logfile:
            logfile.write(info + '\n')
        if stdout:
            print(info)

    def raise_required_param_error(self, parameter):
        raise IOError('Required input parameter "{0}" '
                      'is not defined.'.format(parameter))


    def raise_input_param_error(self, parameter, complaint, explanation):
        raise ValueError('Input parameter "{0}" is not {1}\n '
                      '{2}'.format(parameter, complaint, explanation))

    def report_warning(self, info, stdout=False):
        warning = 'warning: {0}'.format(info)
        self.report2log(warning, stdout=stdout)


def redirectOutput(mode, outputname):
    if mode == 'start':
        config.stdout = os.dup(sys.stdout.fileno())
        config.stdout_file = open(outputname, 'w')
        os.dup2(config.stdout_file.fileno(), sys.stdout.fileno())
    elif mode == 'stop':
        if not config.stdout:
            raise Exception('Output redirection has not been started, '
                            'thus it can not be ended.')
        sys.stdout.flush()
        os.dup2(config.stdout, sys.stdout.fileno())
        config.stdout_file.close()


def redirectErr(mode, outputname):
    if mode == 'start':
        config.stderr = os.dup(sys.stderr.fileno())
        config.stderr_file = open(outputname, 'w')
        os.dup2(config.stderr_file.fileno(), sys.stderr.fileno())
    elif mode == 'stop':
        if not config.stderr:
            raise Exception('Error redirection has not been started, '
                            'thus it can not be ended.')
        sys.stderr.flush()
        os.dup2(config.stderr, sys.stderr.fileno())
        config.stderr_file.close()


def checkDelPhiErrors(filename, mode=None):
    exceptions = []
    if mode == 'readFiles':
        exceptions = ['part of system outside the box!',
                      'has a net charge of']
    elif mode == 'runDelPhi':
        exceptions = ['part of system outside the box!']

    exit_trigger = False
    errors = ''
    with open(filename) as f:
        for line in f:
            ignore_error = False
            if 'WARNING' in line or 'Warning' in line:
                for exception in exceptions:
                    if exception in line:
                        ignore_error = True
                if not ignore_error:
                    exit_trigger = True
                    errors += line
    if exit_trigger:
        raise Exception('The following errors have been found on {0}: \n{1}'.format(filename, errors))
    if not Config.debug:
        if os.path.isfile(filename):
            os.remove(filename)
        focusing_log = '{}_focusing'.format(filename)
        if os.path.isfile(focusing_log):
            os.remove(focusing_log)



import os
from constants import *
from psutil import cpu_count
from pprint import pformat
from numpy import arange
from copy import copy

class Config:
    @classmethod
    def storeParams(cls, titration_obj, log, debug, parameters):
        cls.log = log
        cls.debug = debug
        cls.titration = titration_obj
        cls.pypka_params = PypKaConfig(log)
        cls.delphi_params = DelPhiConfig(log)
        cls.mc_params = MCConfig(log)
        cls.parallel_params = ParallelConfig()

        cls.filter_params(parameters)

        return cls.pypka_params, cls.delphi_params, cls.mc_params, cls.parallel_params

    @classmethod
    def loadParams(cls, parameters):
        pypka_params, delphi_params, mc_params, parallel_params = parameters
        cls.pypka_params = pypka_params
        cls.delphi_params = delphi_params
        cls.mc_params = mc_params
        cls.parallel_params = parallel_params

    @classmethod
    def getConfigObj(cls, name):
        if name in cls.__dict__:
            return cls.__dict__
        elif name in cls.pypka_params:
            return cls.pypka_params
        elif name in cls.delphi_params:
            return cls.delphi_params
        elif name in cls.mc_params:
            return cls.mc_params
        else:
            return None

    @classmethod
    def filter_params(cls, parameters):
        """Checks the validity of the input parameters

        Args:
            parameters (dict): input parameters
        """
        # Define all parameters
        for param_name, param_value in parameters.items():
            if param_name.startswith('sites') or param_name in IGNORED_PARAMS:
                continue
            config_obj = cls.getConfigObj(param_name)
            if config_obj:
                config_obj[param_name] = param_value
            else:
                info = '{} is not a valid parameter.'.format(param_name)
                cls.log.report_warning(info, stdout=True)

        # Check if mandatory parameters were defined
        for param_name in MANDATORY_PARAMS:
            config_obj = cls.getConfigObj(param_name)
            if not (config_obj and config_obj[param_name] is not None):
                cls.log.raise_required_param_error(param_name)

        cls.pypka_params.set_structure_extension()
        cls.pypka_params.set_ncpus()
        cls.pypka_params.set_radii_charges_paths()
        cls.pypka_params.readTermini()

        cls.mc_params.set_pH_values(parameters)

        if cls.pypka_params['structure_output']:
            cls.pypka_params.set_structure_output(cls.mc_params['pHmin'],
                                                   cls.mc_params['pHmax'])

        if cls.delphi_params['pbc_dim'] == 2:
            cls.delphi_params.set_nonlinear_params(cls.pypka_params, parameters)

        if 'lipid_definition' in parameters:
            cls.pypka_params.define_lipids(parameters['lipid_definition'])


class ParametersDict:
    input_conversion = {'structure': 'f_in',
                        'grid_fill': 'perfil',
                        'pbc_dimensions': 'pbc_dim',
                        'convergence': 'maxc',
                        'output': 'f_out',
                        'logfile': 'f_log',
                        'titration_output': 'f_prot_out',
                        'clean': 'clean_pdb'}

    def __init__(self, log):
        self.log = log
        self.input_special_conditions = {}
        self.input_type = {}

    def convert_param_name(self, name):
        if name in self.input_conversion.keys():
            return self.input_conversion[name]
        return name

    def check_param_type(self, param_name, param_value, param_type, msg=''):
        """Checks if param_value is of type param_type

        If param_value is not of param_type, it tries to convert into the desired type.
        An error is raised if the conversion is not possible.

        Args:
            param_name (str): name of the parameter
            param_value: value of the parameter
            param_type (type): desired type for the parameter
            msg (str): error help message

        Returns:
            param_type: param_value is returned with the type param_type
        """
        if param_type is bool and not isinstance(param_value, bool):
            param_value = param_value.lower()
            if param_value in ('false', 'no'):
                param_value = False
            elif param_value in ('true', 'yes'):
                param_value = True

        if not isinstance(param_value, param_type):
            try:
                param_value = param_type(param_value)
            except ValueError:
                self.log.raise_input_param_error(param_name, param_type, msg)
        return param_value

    def check_conditions(self, param_name, param_value):
        if param_name in self.input_type:
            param_type = self.input_type[param_name]
            param_value = self.check_param_type(param_name, param_value, param_type)
        else:
            info = 'parameter {} is not being checked for type. ' \
                    'Please warn the developement team.'.format(param_name)
            self.log.report_warning(info, stdout=True)
        if param_name in self.input_special_conditions:
            condition = self.input_special_conditions[param_name]
            if condition == '>0':
                if  param_value <= 0:
                    self.log.raise_input_param_error(param_name,
                                                     'greater than zero.', '')
            elif param_value not in condition:
                    self.log.raise_input_param_error(param_name,
                                                     'in {}'.format(condition), '')

        return param_value

    def __getitem__(self, name):
        name = self.convert_param_name(name)
        if name not in self.__dict__:
            raise Exception('{} not in {}'.format(name, self.name))
        return self.__dict__[name]

    def __setitem__(self, name, key):
        name = self.convert_param_name(name)
        key = self.check_conditions(name, key)

        self.__dict__[name] = key

    def __contains__(self, item):
        item = self.convert_param_name(item)
        return item in self.__dict__

    def __str__(self):
        self_dict = {}
        for key, value in self.__dict__.items():
            if key not in self.not_to_print and key not in ('name', 'not_to_print', 'log'):
                self_dict[key] = value
        return '# {} Parameters\n{}\n'.format(self.name, pformat(self_dict))


class PypKaConfig(ParametersDict):
    """Configuration parameters
    """
    def __init__(self, log):
        super().__init__(log)

        self.name = 'PypKa'
        self.not_to_print = ['tmpsites', 'temp', 'input_special_conditions', 'input_type', 'NTR_atoms', 'CTR_atoms', 'box']

        self.tmpsites    = {}
        self.pid         = os.getpid()
        self.debug       = False
        self.ncpus       = None
        self.temp        = 298

        self.CpHMD_mode = False

        # Paths
        self.file_dir   = os.path.dirname(os.path.abspath(__file__))
        self.script_dir = os.path.dirname(__file__)
        self.pdb2pqr    = "{0}/pdb2pqr/pdb2pqr.py".format(self.file_dir)
        self.userff     = "{0}/pdb2pqr/dat/GROMOS.DAT".format(self.file_dir)
        self.usernames  = "{0}/pdb2pqr/dat/GROMOS.names".format(self.file_dir)

        # File Naming
        self.f_in               = None
        self.f_in_extension     = None
        self.f_out              = None
        self.f_prot_out         = None
        self.f_structure_out    = None

        # Output File
        self.structure_output   = None
        self.f_structure_out_pH = None
        self.ff_structure_out   = None


        # Force Field
        self.f_crg = None
        self.f_siz = None
        self.ffID  = 'G54A7'
        self.ff_family = 'GROMOS'
        self.NTR_atoms = None
        self.CTR_atoms = None
        self.LIPIDS = {}

        # Preprocessing parameters
        self.ffinput    = 'GROMOS'
        self.clean_pdb  = True
        self.keep_ions  = False
        self.ser_thr_titration = True

        self.cutoff     = -1
        self.slice      = 0.05
        self.box        = []

        # Parameters Validity
        self.input_special_conditions = {
            'temp': '>0',
            'ffID': ('G54A7', 'CHARMM36m'),
            'ffinput': ('GROMOS', 'AMBER', 'CHARMM'),
            'ff_structure_out': ('gromos_cph', 'amber')
        }
        self.input_type = {
            'debug'             : bool,
            'ncpus'             : int,
            'temp'              : float,
            'f_in'              : str,
            'f_out'             : str,
            'f_prot_out'        : str,
            'f_structure_out'   : str,
            'f_structure_out_pH': float,
            'ff_structure_out'  : str,
            'structure_output'  : str,
            'ffID'              : str,
            'ff_family'         : str,
            'ffinput'           : str,
            'cutoff'            : float,
            'slice'             : float,
            'clean_pdb'         : bool,
            'keep_ions'         : bool,
            'ser_thr_titration' : bool,
            'f_crg'             : str,
            'f_siz'             : str,
            'box'               : list,
            'CpHMD_mode'        : bool
        }

    def set_structure_extension(self):
        structure = self['structure']
        f_in_parts = structure.split('.')
        if len(f_in_parts) <= 1:
            self.log.raise_input_param_error('structure',
                'a string containing a file extension.',
                'Ex: structure.pdb or structure.gro')

        extension = f_in_parts[-1].lower().replace('pqr', 'pdb')
        if extension not in ('gro', 'pdb'):
            self.log.raise_input_param_error('structure',
                'a string containing a valid file extension.',
                'Ex: structure.pdb or structure.gro or structure.pqr')

        self.f_in_extension = extension

    def setBox(self, box):
        self['box'] = box

    def set_ncpus(self):
        ncpus = self['ncpus']
        if ncpus == -1:
            self['ncpus'] = cpu_count(logical=False)
        elif ncpus < 1:
            self['ncpus'] = 1

    def set_radii_charges_paths(self):
        file_path = os.path.join(self['script_dir'], self['ffID'])
        if not self['f_crg']:
            self['f_crg'] = '{}/DataBaseT.crg'.format(file_path)
        if not self['f_siz']:
            self['f_siz'] = '{}/DataBaseT.siz'.format(file_path)

        ffID = Config.pypka_params['ffID'].lower()
        if 'charmm36m' in ffID:
            self['ff_family'] = 'CHARMM'
        elif 'g54a7' in ffID:
            self['ff_family'] = 'GROMOS'
        else:
            raise Exception('Forcefield {0} not supported'.format(ffID))


    def define_lipids(self, lipids):
        for lipid in lipids:
            resname = lipids[lipid]
            self.LIPIDS[lipid] = resname

    def set_structure_output(self, pHmin, pHmax):
        error_raise = False
        structure_output = self['structure_output'].split(',')
        msg = 'CLI Example: "structure_output": ("structure.pdb", 7, "amber")\n '\
              'API Example: structure_output = structure.pdb, 7, amber'
        if len(structure_output) != 3:
            error_msg = 'a tuple containing a filename, the desired pH value and force field naming scheme.'
            self.log.raise_input_param_error('structure_output', error_msg, msg)


        outfilename = structure_output[0].strip('()"\' ')
        pH =  structure_output[1].strip('()"\' ')
        ff_out = structure_output[2].strip('()"\' ').lower()

        pH = self.check_param_type('structure_output_pH', pH, float, msg)

        self['f_structure_out'] = outfilename
        self['f_structure_out_pH'] = pH
        self['ff_structure_out'] = ff_out

        if pH < pHmin or pH > pHmax:
            message = 'in range [pHmin, pHmax].'
            self.log.raise_input_param_error('structure_output', message, '')

    def readTermini(self):
        script_dir = self['script_dir']
        ffID = self['ffID']
        NTR_atoms = []
        ntr_fname = "{}/{}/sts/NTRtau1.st".format(script_dir, ffID)
        with open(ntr_fname) as f:
            for line in f:
                parts = line.split()
                if len(parts) > 1:
                    NTR_atoms.append(parts[1].strip())
        CTR_atoms = []
        ctr_fname = "{}/{}/sts/CTRtau1.st".format(script_dir, ffID)
        with open(ctr_fname) as f:
            for line in f:
                parts = line.split()
                if len(parts) > 1:
                    CTR_atoms.append(parts[1].strip())
        self.NTR_atoms = NTR_atoms
        self.CTR_atoms = CTR_atoms

    def redefine_f_in(self, new_f_in):
        self['f_in'] = new_f_in
        self.set_structure_extension()


class DelPhiConfig(ParametersDict):
    """DelPhi configuration parameters
    """
    def __init__(self, log):
        super().__init__(log)

        self.name = 'DelPhi'
        self.not_to_print = ['p_atpos', 'p_rad3', 'p_chrgv4', 'atinf', 'p_iatmed',
                             'delphimol', 'input_type', 'input_special_conditions', 'lookup_atoms']

        self.perfil     = 0.9
        self.gsize      = 81
        self.scaleP     = 1
        self.scaleM     = 4
        self.precision  = 'single'
        self.ionicstr   = 0.1
        self.bndcon     = 3
        self.maxc       = 0.01
        self.nlit       = 500
        self.nonit      = 0
        self.relfac     = 0.75
        self.relpar     = 0.75
        self.pbx        = False
        self.pby        = False
        self.epssol     = 80.0
        self.pbc_dim    = 0
        self.epsin      = 20.0

        self.lookup_atoms = None

        # -1 NanoShaper off
        # 0 connolly surface
        # 1 skin
        # 2 blobby
        # 3 mesh
        # 4 msms
        self.nanoshaper = -1

        self.p_atpos   = None
        self.p_rad3    = None
        self.p_chrgv4  = None
        self.atinf     = None
        self.p_iatmed  = None
        self.delphimol = None

        self.input_type = {
            'perfil'    : float,
            'gsize'     : int,
            'scaleP'    : float,
            'scaleM'    : float,
            'precision' : str,
            'ionicstr'  : float,
            'bndcon'    : int,
            'maxc'      : float,
            'nlit'      : int,
            'nonit'     : int,
            'relfac'    : float,
            'relpar'    : float,
            'pbx'       : bool,
            'pby'       : bool,
            'epssol'    : float,
            'pbc_dim'   : int,
            'epsin'     : float,
            'nanoshaper': int
        }

        self.input_special_conditions = {
            'scaleP'   : '>0',
            'scaleM'   : '>0',
            'maxc'     : '>0',
            'gsize'    : '>0',
            'perfil'   : '>0',
            'ionicstr' : '>0',
            'nlit'     : '>0',
            'epssol'   : '>0',
            'epsin'    : '>0',
            'bndcon'   : (1, 2, 3, 4),
            'precision': ('single', 'double'),
            'pbc_dim'  : (0, 2),
            'nanoshaper': (0, 1, 2, 3, 4, -1)
        }

    def set_nonlinear_params(self, pypka_config, input_params):
        if self['relfac'] != 0.2 and \
            'relfac' not in input_params:
            self['relfac'] = 0.2
        if self['nonit'] != 5 and \
            'nonit' not in input_params:
            self['nonit'] = 5
        if pypka_config['cutoff'] == -1 and \
            'cutoff' not in input_params:
            pypka_config['cutoff'] = 5

    def redefineScale(self):
        scaleP = (self['gsize'] - 1) / (Config.pypka_params['box'][0])
        scaleM = int(4 / scaleP + 0.5) * scaleP

        self['scaleP'] = scaleP
        self['scaleM'] = scaleM

    def store_run_params(self, delphimol):
        if delphimol != 'reload':
            self.delphimol = delphimol
        else:
            delphimol = self.delphimol
        self.p_atpos  = copy(delphimol.get_atpos())
        self.p_rad3   = copy(delphimol.get_rad3())
        self.p_chrgv4 = copy(delphimol.get_chrgv4())
        self.atinf    = copy(delphimol.get_atinf())
        self.p_iatmed = copy(delphimol.get_iatmed())


class MCConfig(ParametersDict):
    """Monte Carlo configuration parameters
    """
    def __init__(self, log):
        super().__init__(log)

        self.name = 'Monte Carlo'
        self.not_to_print = ['input_type', 'input_special_conditions']
        self.pHmin      = 0
        self.pHmax      = 14
        self.pHstep     = 0.25
        self.seed       = 1234567
        self.couple_min = 2.0
        self.mcsteps    = 200000
        self.eqsteps    = 1000

        self.pH_values = []

        self.input_special_conditions = {
            'pHstep' : '>0',
            'mcsteps': '>0',
            'eqsteps': '>0'
        }
        self.input_type = {
            'pHmin'     : float,
            'pHmax'     : float,
            'pHstep'    : float,
            'seed'      : int,
            'couple_min': float,
            'mcsteps'   : int,
            'eqsteps'   : int
        }

    def set_pH_values(self, parameters):
        if 'pH' in parameters:
            pH = parameters['pH']
        else:
            pH = '{}-{}'.format(self.pHmin, self.pHmax)

        pH_error_info = 'pH can be a single value or a range. '\
                        'As default pH is set to [0, 14]\n'\
                        'API Example: pH = 7 or pH = 5,8\n'\
                        'CLI Example: "pH": 7 or "pH": "5,18"'

        pH_parts = pH.split(',')
        if len(pH_parts) == 1:
            pH_parts = pH.split('-')
        if len(pH_parts) == 2:
            self.pHmin = self.check_param_type('pHmin', pH_parts[0], float)
            self.pHmax = self.check_param_type('pHmax', pH_parts[1], float)

            diff = self.pHmax - self.pHmin
            if diff < 0:
                self.log.raise_input_param_error('pH', 'correctly defined.',
                                                 pH_error_info)
            if diff % self.pHstep != 0:
                self.pHmax += self.pHstep
            self.pH_values = arange(self.pHmin, self.pHmax + 0.001, self.pHstep)
            self.pHmax = float(self.pH_values[-1])
        elif len(pH_parts) == 1:
            pH = self.check_param_type('pH', pH, float)
            self.pH_values = (pH)
            self.pHmin = pH
            self.pHmax = pH
        else:
            self.log.raise_input_param_error('pH', 'correctly defined.',
                                             pH_error_info)


class ParallelConfig:
    def __init__(self):
        self.njobs      = None
        self.pb_time    = None
        self.total_jobs = None

        self.all_tautomers_order = None
        self.site_interactions = None
        self.interactions_look = None
        self.all_sites = None

        self.npossible_states = None
        self.possible_states_g = None
        self.possible_states_occ = None
        self.interactions = None



import setuptools

long_description = """
# PypKa

A python module for flexible Poisson-Boltzmann based pKa calculations with proton tautomerism


# Dependencies

  - libgfortran4
  - gawk
  - pytest
  - numpy
  - psutil


# License

  pypka is distributed under a LGPL-3.0, however delphi4py depends on
  DelPhi which is proprietary. To use DelPhi the user is required to
  download the DelPhi license
  [here](https://honiglab.c2b2.columbia.edu/software/cgi-bin/software.pl?input=DelPhi)

# Documentation

  Documentation can be found [here](https://pypka.readthedocs.io/en/latest/). (Under development)

# Installation

  pip3 install pypka

  apt install gawk gcc gfortran libgfortran4

# Contacts

  Please submit a github issue to report bugs and to request new features.
  Alternatively you may find the developer [here](mailto:pdreis@fc.ul.pt). Please visit ou [website](http://mms.rd.ciencias.ulisboa.pt/) for more information.

"""



setuptools.setup(
    name="pypka",
    version="1.2.0",
    author="Pedro Reis",
    author_email="pdreis@fc.ul.pt",
    description="A python module for flexible Poisson-Boltzmann based pKa calculations with proton tautomerism",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mms-fcul/PypKa",
    packages=setuptools.find_packages(),
    package_dir={'pypka': 'pypka'},
    package_data={'pypka': ['mc*.so',
                            'addHtaut*',
                            'G54A7/*',
                            'G54A7/sts/*',
                            'pdb2pqr/*',
                            'pdb2pqr/dat/*',
                            'pdb2pqr/extensions/*',
                            'pdb2pqr/src/*',
                            'pdb2pqr/ZSI/*',
                            'delphi4py/*',
                            'delphi4py/readFiles/*.so',
                            'delphi4py/rundelphi/*.so']},
    install_requires=['numpy', 'pytest', 'coverage', 'psutil'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3) ",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    entry_points={
        'console_scripts': [
            'pypka = pypka.pypka:CLI'
        ]
    }
)



import sys
import os
sys.path.insert(1, os.getcwd() +  '/../../')

from formats import read_pdb_line

def storeResidues(filename):
    residues = {}
    with open(filename) as f_original:
        for line in f_original:
            if line.startswith('ATOM '):
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)
                if resnumb not in residues:
                    residues[resnumb] = {}
                residues[resnumb][aname] = (resname, x, y, z)
    return residues

def roundXYZ(x, y, z):
    x = round(x, 2)
    y = round(y, 2)
    z = round(z, 2)
    return x, y, z

def areCloseEnough(p1, p2):
    x1, y1, z1 = p1
    x2, y2, z2 = p2
    return (abs(x1 - x2) <= 0.1 and 
            abs(y1 - y2) <= 0.1 and
            abs(z1 - z2) <= 0.1)

def dist3D(p1, p2):
    x1, y1, z1 = p1
    x2, y2, z2 = p2    
    
    return ((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2) ** 0.5


def compareFiles(f1, f2):
    original = storeResidues(f1)
    new = storeResidues(f2)

    assert(len(original) == len(new))

    problems = False
    
    for resid in original:
        res = original[resid]
        resnew = new[resid]
        len_orig = len(res)
        len_new = len(resnew)
        if len_orig != len_new:
            print(f'Error in res #{resid} {f1}: {len_orig} {f2}: {len_new}')
            problems = True
            for i in res:
                if i not in resnew:
                    print(f'{i} not in {f2}')
            for i in resnew:
                if i not in res:
                    print(f'{i} not in {f1}')

        for atom in res:
            if atom not in resnew:
                print(f'{atom} not in {res[atom][0]}{resid} {f2}')
                problems = True
                continue
            (resname1, x1, y1, z1) = res[atom]
            (resname2, x2, y2, z2) = resnew[atom]
            if (resname1 != resname2):
                print(f'Warning: resname {resname1} != {resname2}')
                problems = True
            p1 = roundXYZ(x1, y1, z1)
            p2 = (x2, y2, z2)
            if not areCloseEnough(p1, p2):
                WarningTrigger = True
                for atom2 in resnew:
                    (resname2, x2, y2, z2) = resnew[atom2]
                    p2 = (x2, y2, z2)
                    if areCloseEnough(p1, p2):
                        WarningTrigger = False
                        break
                if WarningTrigger:
                    print(f'{resnew[atom][0]}{resid} {atom}: {p1} != {resnew[atom][1:]} {dist3D(p1, p2)}')

        for atom in resnew:
            if atom not in res:
                print(f'{atom} not in {f2} {resnew[atom][0]}{resid}')
                problems = True
                continue
    return problems
